/*
    Ishtar - A network transparent read/write interface with
    probing and enumeration support.
    network part : convenient cross-platform packaging for tcp/ip networking
    Copyright (C) 2003-2005 Stephane Magnenat <nct@ysagoon.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __ISHTAR_STREAM_H
#define __ISHTAR_STREAM_H

#include <algorithm>

/*!	\file stream.h
	\brief Interface of stream
*/

/*!	\mainpage Ishtar network library reference documentation

	\section intro Introduction
	Ishtar is a network toolbox.
	Ishtar is composed of three different modules.
	This modules is libishtarnet, the most basic one. It provides clean, endian-safe and easy to use TCP/IP sockets wrapper. They are based on a stream interface. A filesystem implementation of this interface is also provided along with the network one.
	
	\section installation Installation
	The installation procedure is explained step by step in the \ref InstallGuide.
	
	\section usage Usage
	A quick start to most important classes is given in the \ref Cookbook.
	
	\section fileformat File Format
	Basic data types such as int, float and double are serialized in little endian. Although the so-called net-order is big endian, most used architectures are using little endian. Conversion is transparently performed on big endian platforms.
	
	\section feedback Feedback
	If you have any comments or suggestions, do not hesitate to send them to
	Stephane Magnenat <nct@ysagoon.com>
	
	\section license License
	This program is free software released under the GNU General Public License version 2.
*/

//! Ishtar is the namespace of the Ishtar class library
namespace Ishtar
{
	//! 8 bits signed integer
	typedef signed char Int8;
	//! 8 bits unsigned integer
	typedef unsigned char UInt8;
	//! 16 bits signed integer
	typedef signed short Int16;
	//! 16 bits unsigned integer
	typedef unsigned short UInt16;
	//! 32 bits signed integer
	typedef signed int Int32;
	//! 32 bits unsigned integer
	typedef unsigned int UInt32;
	//! unsigned integer, size big enough to represents as many elements as there is bytes in memory. Same size as pointer
	typedef size_t Size;
	
	//! In constructor, discover the endianess
	struct EndianDiscoverer
	{
		//! True if system is big endian
		bool isLittleEndian;
		//! True if double type has inverted words
		bool isInvertedDouble;
		
		//! Constructor, test for byte order and set isBigEndian
		EndianDiscoverer();
	};
	
	//! Static instance of endianDiscoverer, can be queried to know what the system endian is
	extern EndianDiscoverer endianDiscoverer;
	
	//! Swap bytes if system is little endian
	template<typename T> void swapBytesIfRequired(T &v)
	{
		// we save files in little endian as most user architecture are little endian
		if (!endianDiscoverer.isLittleEndian)
		{
			UInt8 *data = reinterpret_cast<UInt8 *>(&v);
			Size size = sizeof(T);
			for(Size p = 0; p < (size>>1); p++)
				std::swap(data[p], data[size-1-p]);
		}
	}
	
	// Endian safe read and write
	// ==========================
	
	//! An output stream to which we can write
	class OutputStream
	{
	public:
		//! Destructor
		virtual ~OutputStream() { }
	
		//! Write a data buffer, no endian conversion
		virtual void write(const void *data, const Size size) = 0;
	
		//! Flush the internal buffer, be sure that datas are really written
		virtual void flush(void) = 0;
	
		//! Write buffer v, correct endian
		template<typename T> void writeEndianIndependant(const T &v)
		{
			T swapedValue = v;
			swapBytesIfRequired(swapedValue);
			write(&swapedValue, sizeof(T));
		}
	
		//! Write a 8 bits signed integer, endian safe
		inline void writeInt8(const Int8 v) { this->write(&v, 1); }
		//! Write a 8 bits unsigned integer, endian safe
		inline void writeUInt8(const UInt8 v) { this->write(&v, 1); }
		//! Write a 16 bits signed integer, endian safe
		inline void writeInt16(const Int16 v) { this->writeEndianIndependant(v); }
		//! Write a 16 bits unsigned integer, endian safe
		inline void writeUInt16(const UInt16 v) { this->writeEndianIndependant(v); }
		//! Write a 32 bits signed integer, endian safe
		inline void writeInt32(const Int32 v) { this->writeEndianIndependant(v); }
		//! Write a 32 bits unsigned integer, endian safe
		inline void writeUInt32(const UInt32 v) { this->writeEndianIndependant(v); }
		//! Write a float, endian safe
		inline void writeFloat(const float v) { this->writeEndianIndependant(v); }
		//! Write a double, endian safe
		inline void writeDouble(const double v)
		{
			double value = v;
			// arm inverted double forces us to do this
			if (endianDiscoverer.isInvertedDouble)
			{
				UInt32 *ptr = reinterpret_cast<UInt32 *>(&value);
				UInt32 tempValue = *ptr;
				*ptr = *(ptr+1);
				*(ptr+1) = tempValue;
			}
			this->writeEndianIndependant(value);
		}
	};
	
	//! An input stream from which we can read
	class InputStream
	{
	public:
		//! Destructor
		virtual ~InputStream() { }
	
		//! Read a data buffer, no endian conversion
		virtual void read(void *data, Size size) = 0;
	
		//! Read buffer v, correct endian for size 2, 4, and 8 (2*4)
		template<typename T> T readEndianIndependant()
		{
			T v;
			read(&v, sizeof(T));
			swapBytesIfRequired(v);
			return v;
		}
	
		//! Read a 8 bits signed integer, endian safe
		inline Int8 readInt8(void) { Int8 i; this->read(&i, 1); return i; }
		//! Read a 8 bits unsigned integer, endian safe
		inline UInt8 readUInt8(void) { UInt8 i; this->read(&i, 1); return i; }
		//! Read a 16 bits signed integer, endian safe
		inline Int16 readInt16(void) { return this->readEndianIndependant<Int16>(); }
		//! Read a 16 bits unsigned integer, endian safe
		inline UInt16 readUInt16(void) { return this->readEndianIndependant<UInt16>(); }
		//! Read a 32 bits signed integer, endian safe
		inline Int32 readInt32(void) { return this->readEndianIndependant<Int32>(); }
		//! Read a 32 bits unsigned integer, endian safe
		inline UInt32 readUInt32(void) { return this->readEndianIndependant<UInt32>(); }
		//! Read a float, endian safe
		inline float readFloat(void) { return this->readEndianIndependant<float>(); }
		//! Read a double, endian safe
		inline double readDouble(void)
		{
			double value = this->readEndianIndependant<double>();
			// arm inverted double forces us to do this
			if (endianDiscoverer.isInvertedDouble)
			{
				UInt32 *ptr = reinterpret_cast<UInt32 *>(&value);
				UInt32 tempValue = *ptr;
				*ptr = *(ptr+1);
				*(ptr+1) = tempValue;
			}
			return value;
		}
	};
}

#endif
